from bson import ObjectId
from datetime import datetime
from dataclasses import dataclass, field


@dataclass(slots=True)
class DecisionMaker:
    # Foreign key to the `company` collection
    company_id: str
    firstname: str
    lastname: str
    source: str
    # The raw string from the source, will cleaned and classified into `function_string`
    raw_function_string: str
    last_seen: datetime
    _id: ObjectId = field(default_factory=ObjectId)
    function_string: str | None = None
    email: str | None = None
    email_score: int | None = None
    phone: str | None = None
    phone_score: int | None = None
    language: str | None = None
    function_code: int | None = None
    linkedin_url: str | None = None
    cluster: str | None = None
    cluster_score: float | None = None
    cluster_best_match: str | None = None


@dataclass(slots=True)
class DecisionMakerLegacy:
    """
    Decision makers imported from the legacy SQL database.

    This is deperecated and kept for reference purpose only.
    """

    # Company ID (same _id as in company collection with leading zeros BE:012345667)
    company_id: str
    # Company ID (id in the old format (VAT without leading zeros) BE:12345667)
    legacy_entity_id: str
    name: str | None = None
    firstname: str | None = None
    # deprecated
    language: str | None = None
    # Gender (deprecated)
    sex: str | None = None
    # deprecated
    function: str | None = None
    source: str | None = None
    function_string: str | None = None
    linkedin_url: str | None = None
    email: str | None = None
    email_source: str | None = None
    email_score: int | None = None
    blacklisted: str | None = None
    # deprecated
    phone: str | None = None
    cluster: str | None = None
    cluster_score: int | None = None
    # cluster match
    best_match: str | None = None
    position_request_base: int | None = None
    position_request_ceo: int | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    deleted_at: datetime | None = None
    google_last_seen: datetime | None = None
    norbert_last_seen: datetime | None = None



# For local dev purpose only
from dotenv import load_dotenv

load_dotenv()
import json
import time
import traceback
from typing import Callable
from requests import Session
from functools import partial
from queue import Queue, Empty
from threading import Thread, Event
from concurrent.futures import ThreadPoolExecutor

from inoopa_utils.inoopa_logging import create_logger
from inoopa_utils.mongodb_helpers import DbManagerMongo
from inoopa_utils.rabbitmq_helpers import get_messages_from_queue, push_to_queue


def parralelize_company_processing_and_saving(
    processing_function: Callable,
    input_queue_name: str,
    process_name: str,
    processing_function_need_session: bool = False,
    pretty_logging: bool = False,
) -> None:
    """
    Process the queue of companies to process and save them in DB.
    It can seems overengineered but we went from 20 days to 2h to scrape the whole BCE database with this.

    In this function, 2 groups of threads are running in parallel:
        - Multiple threads for processing
        - 1 thread for saving in DB

    :param processing_function: function to use to process the companies. Any function used here should these params:
        - `params` as a parameter. This is the dict containing the company params to process. This can be nammed as you want.
        - `queue_name` as a parameter. This is the name of the queue to push the companies to requeue
            or to deduce the dead letter queue name in case of erroers.
        - Optional: `session` as a parameter. This is only needed if the processing function makes a lot of requests.
    :param input_queue_name: name of the queue containing the companies to process with it's params
    :param process_name: name of the process function (for logging purpose) ex: bce_scraper, website_finder,...
    :param processing_function_need_session: whether the processing function needs a session or not.
        When the processing function makes a lot of requests, it is better to use a session
        to avoid creating a new connection for each request. (your function should have a `session` as a parameter)
    """
    logger = create_logger(f"INOOPA.ETL.{process_name.upper()}.BATCH_PROCESSING", pretty=pretty_logging)
    dead_letter_queue_name = f"{input_queue_name}_DLQ"
    # This queue is processed in a separate thread to avoid blocking the main thread (processing) when saving in DB
    db_queue = Queue()
    # This event will be used to send an exit signal to the DB thread
    db_thread_exit_signal = Event()
    # This thread will consume the db_queue and save companies in DB in a separate thread
    # This is done to avoid blocking the main thread (scraping) when saving in DB (which can take a while)
    db_manager_thread = Thread(target=_save_companies_in_db, args=(db_queue, db_thread_exit_signal), daemon=True)
    db_manager_thread.start()

    while True:
        # This queue's messages should only contains a list of companies number
        queue_message = get_messages_from_queue(input_queue_name)
        if not queue_message:
            logger.info("No more companies to process, stop...")
            # Send the exit signal to the DB thread
            db_thread_exit_signal.set()
            break

        try:
            # This will load a list of dict from the queue message. Each dict represent one company to process.
            # This dict should contains the keys the processing function needs.
            # The bce_scraper only needs the company number for example.
            # So the queue messages should contain a LIST of DICT like this: [{"company_number": "0649.973.640"},...]
            processing_function_params = json.loads(queue_message[0])
            logger.info(f"Processing {len(processing_function_params)} companies...")
            if processing_function_need_session:
                with Session() as session:
                    with ThreadPoolExecutor() as executor:
                        processed_companies = list(
                            executor.map(
                                partial(processing_function, queue_name=input_queue_name, session=session),
                                processing_function_params,
                            )
                        )
            else:
                with ThreadPoolExecutor() as executor:
                    processed_companies = list(
                        executor.map(
                            partial(processing_function, queue_name=input_queue_name),
                            processing_function_params,
                        )
                    )
            # Add companies to the python queue to be saved in DB
            if processed_companies:
                db_queue.put(processed_companies)
        except Exception as ex:
            logger.error(
                f"Error: {ex} while processing batch, sending batch to dead-letter-queue: {dead_letter_queue_name}"
            )
            # print traceback in case of error for debugging purpose
            traceback.print_exc()
            push_to_queue(dead_letter_queue_name, processing_function_params)
            # Send the exit signal to the DB thread
            db_thread_exit_signal.set()
            # Wait for the DB thread to finish before exiting (Wait for all the companies to be saved in DB)
            db_manager_thread.join()
            raise ex
    # Wait for the DB thread to finish before exiting (Wait for all the companies to be saved in DB)
    db_manager_thread.join()


def _save_companies_in_db(db_python_queue: Queue, exit_signal: Event) -> None:
    """
    Save companies in DB.

    :param db_python_queue: queue containing the companies to save in DB
    :param exit_signal: event to send an exit signal to the thread
    """
    _logger = create_logger("INOOPA.BATCH_PROCESSING.DB_THREAD")
    db_manager = DbManagerMongo()
    # Wait for the python queue to be empty and the exit signal to be set to end the thread
    while True:
        try:
            # A batch of companies is pulled from the python queue.
            # Pull during 1 second max to avoid blocking the thread. DO NOT REMOVE THE TIMEOUT!
            companies = db_python_queue.get(timeout=1)
            db_manager.update_or_add_many_to_collection(companies)
        except Empty:
            if exit_signal.is_set():
                break
            time.sleep(1)
    _logger.info("DB thread exiting...")



from datetime import datetime
from typing import Literal

from pydantic import BaseModel


MongoFilters = dict[str, dict[str, dict[str, int | str | list[str]]]]

AdditionalFields = Literal["email", "phone", "website", "nace_codes", "social_medias"]

Country = Literal["BE", "FR", "NL"]
all_countries: list[Country] = ["BE", "FR", "NL"]

Region = Literal["bruxelles", "flamande", "wallonne", "NOT FOUND"]
all_regions: list[Region] = ["bruxelles", "flamande", "wallonne", "NOT FOUND"]

Province = Literal[
    "antwerpen",
    "brabant flamand",
    "brabant wallon",
    "bruxelles",
    "flandre occidentale",
    "flandre orientale",
    "hainaut",
    "limbourg",
    "liège",
    "luxembourg",
    "namur",
    "NOT FOUND",
]
all_provinces: list[Province] = [
    "antwerpen",
    "brabant flamand",
    "brabant wallon",
    "bruxelles",
    "flandre occidentale",
    "flandre orientale",
    "hainaut",
    "limbourg",
    "liège",
    "luxembourg",
    "namur",
    "NOT FOUND",
]

LegalFormType = Literal["for-profit", "non-profit", "public"]
all_legal_form_types: list[LegalFormType] = ["for-profit", "non-profit", "public"]

EmployeeCategory = Literal[
    "0 employees",
    "1 to 4 employees",
    "5 to 9 employees",
    "10 to 19 employees",
    "20 to 49 employees",
    "50 to 99 employees",
    "100 to 199 employees",
    "200 to 499 employees",
    "500 to 999 employees",
    "1000 to 9999999 employees",
]
all_employee_categories: list[EmployeeCategory] = [
    "0 employees",
    "1 to 4 employees",
    "5 to 9 employees",
    "10 to 19 employees",
    "20 to 49 employees",
    "50 to 99 employees",
    "100 to 199 employees",
    "200 to 499 employees",
    "500 to 999 employees",
    "1000 to 9999999 employees",
]

LegalForm = Literal[
    "Agricultural company",
    "Autonomous municipal company",
    "Autonomous provincial company",
    "Brussels-Capital region authority",
    "CPAS / OCMW Association",
    "Cities and municipalities",
    "Co-ownership association",
    "Common law company",
    "Company or association without legal personality",
    "Congolese company",
    "Cooperative society",
    "Cooperative society (old regime)",
    "Cooperative society governed by public law",
    "Cooperative society governed by public law (old regime)",
    "Cooperative society with limited liability",
    "Cooperative society with limited liability (profit share)",
    "Cooperative society with limited liability and a social objective",
    "Cooperative society with limited liability governed by public law",
    "Cooperative society with unlimited liability",
    "Cooperative society with unlimited liability (profit share)",
    "Cooperative society with unlimited liability and a social objective",
    "Economic interest grouping with a social objective",
    "Economic interest grouping with registered seat in Belgium",
    "Europ. Econ. assoc wo reg.seat but with est. unit in Belgium",
    "European company (Societas Europaea)",
    "European cooperative society",
    "European economic assoc with registered seat in Belgium",
    "European political foundation",
    "European political party",
    "Federal public planning service",
    "Federal public service",
    "Flemish region and Flemish community authority",
    "Foreign ent. with property in Belgium (without legal pers.)",
    "Foreign entity",
    "Foreign entity with property in Belgium (with legal personality)",
    "Foreign entity without Belgian establishment unit with VAT representation",
    "Foreign listed company without Belgian establishment unit",
    "Foreign or international public organisations",
    "French community authority",
    "General partnership",
    "General partnership with a social objective",
    "German-speaking community authority",
    "Health fund / Mutual health insurance / National union of health funds",
    "Hulpverleningszone",
    "Intercommunal",
    "International non-profit association",
    "International non-profit association governed by public law",
    "International scientific organisation under Belgian law",
    "Limited partnership",
    "Limited partnership governed by public Law",
    "Local police",
    "Ministry for Middle Class",
    "Ministry of Economic Affairs",
    "Ministry of Foreign Affairs",
    "Ministry of Home Affairs",
    "Ministry of Justice",
    "Ministry of the Brussels-Capital Region",
    "Ministry of the Flemish Community",
    "Ministry of the French Community",
    "Ministry of the Walloon Region",
    "Miscellaneous",
    "Miscellaneous without legal personality",
    "Non-profit institution",
    "Non-profit organisation",
    "Ordinary limited partnership",
    "Ordinary limited partnership with a social objective",
    "Organis. regist. with the public admin. Pensions (Finance)",
    "Organisations registered with the O.N.P",
    "Other federal services",
    "Other institution with a social objective (public)",
    "Other legal form",
    "Other private organisation with legal personality",
    "Partnership limited by shares",
    "Partnership limited by shares with a social objective",
    "Pawnshop",
    "Pension scheme organisation",
    "Polders and water boards",
    "Private foreign association with establishment in Belgium",
    "Private foundation",
    "Private limited company",
    "Private limited company governed by public law",
    "Private limited liability company",
    "Private limited liability company with a social objective",
    "Private mutual insurance fund",
    "Professional corporations - Orders",
    "Professional union",
    "Project association",
    "Provincial authority",
    "Pubic social action centre",
    "Public institution",
    "Public limited company",
    "Public limited company with a social objective",
    "Public non-profit association",
    "Public utility foundation",
    "Public utility institution",
    "Representative association (Flemish region)",
    "Service provider association (Flemish region)",
    "State, Province, Region, Community",
    "Temporary association",
    "The services of the Prime Minister",
    "Trade union",
    "Unkown legal form (NSSO)",
    "VAT-group",
    "Walloon region authorit",
]
all_legal_forms: list[LegalForm] = [
    "Agricultural company",
    "Autonomous municipal company",
    "Autonomous provincial company",
    "Brussels-Capital region authority",
    "CPAS / OCMW Association",
    "Cities and municipalities",
    "Co-ownership association",
    "Common law company",
    "Company or association without legal personality",
    "Congolese company",
    "Cooperative society",
    "Cooperative society (old regime)",
    "Cooperative society governed by public law",
    "Cooperative society governed by public law (old regime)",
    "Cooperative society with limited liability",
    "Cooperative society with limited liability (profit share)",
    "Cooperative society with limited liability and a social objective",
    "Cooperative society with limited liability governed by public law",
    "Cooperative society with unlimited liability",
    "Cooperative society with unlimited liability (profit share)",
    "Cooperative society with unlimited liability and a social objective",
    "Economic interest grouping with a social objective",
    "Economic interest grouping with registered seat in Belgium",
    "Europ. Econ. assoc wo reg.seat but with est. unit in Belgium",
    "European company (Societas Europaea)",
    "European cooperative society",
    "European economic assoc with registered seat in Belgium",
    "European political foundation",
    "European political party",
    "Federal public planning service",
    "Federal public service",
    "Flemish region and Flemish community authority",
    "Foreign ent. with property in Belgium (without legal pers.)",
    "Foreign entity",
    "Foreign entity with property in Belgium (with legal personality)",
    "Foreign entity without Belgian establishment unit with VAT representation",
    "Foreign listed company without Belgian establishment unit",
    "Foreign or international public organisations",
    "French community authority",
    "General partnership",
    "General partnership with a social objective",
    "German-speaking community authority",
    "Health fund / Mutual health insurance / National union of health funds",
    "Hulpverleningszone",
    "Intercommunal",
    "International non-profit association",
    "International non-profit association governed by public law",
    "International scientific organisation under Belgian law",
    "Limited partnership",
    "Limited partnership governed by public Law",
    "Local police",
    "Ministry for Middle Class",
    "Ministry of Economic Affairs",
    "Ministry of Foreign Affairs",
    "Ministry of Home Affairs",
    "Ministry of Justice",
    "Ministry of the Brussels-Capital Region",
    "Ministry of the Flemish Community",
    "Ministry of the French Community",
    "Ministry of the Walloon Region",
    "Miscellaneous",
    "Miscellaneous without legal personality",
    "Non-profit institution",
    "Non-profit organisation",
    "Ordinary limited partnership",
    "Ordinary limited partnership with a social objective",
    "Organis. regist. with the public admin. Pensions (Finance)",
    "Organisations registered with the O.N.P",
    "Other federal services",
    "Other institution with a social objective (public)",
    "Other legal form",
    "Other private organisation with legal personality",
    "Partnership limited by shares",
    "Partnership limited by shares with a social objective",
    "Pawnshop",
    "Pension scheme organisation",
    "Polders and water boards",
    "Private foreign association with establishment in Belgium",
    "Private foundation",
    "Private limited company",
    "Private limited company governed by public law",
    "Private limited liability company",
    "Private limited liability company with a social objective",
    "Private mutual insurance fund",
    "Professional corporations - Orders",
    "Professional union",
    "Project association",
    "Provincial authority",
    "Pubic social action centre",
    "Public institution",
    "Public limited company",
    "Public limited company with a social objective",
    "Public non-profit association",
    "Public utility foundation",
    "Public utility institution",
    "Representative association (Flemish region)",
    "Service provider association (Flemish region)",
    "State, Province, Region, Community",
    "Temporary association",
    "The services of the Prime Minister",
    "Trade union",
    "Unkown legal form (NSSO)",
    "VAT-group",
    "Walloon region authorit",
]


class LeadGenerationCompanyFilters(BaseModel):
    """
    Represents the filters used to generate the leads.

    This is a pydantic model, so it can be used in the apis as a parameter.
    """

    additional_fields: list[AdditionalFields] | None = ["email", "phone", "website", "nace_codes", "social_medias"]

    countries: list[Country] | None = None
    regions: list[Region] | None = None
    provinces: list[Province] | None = None
    zipcodes: list[str] | None = None
    declared_best_nace_codes: list[str] | None = None
    declared_best_nace_codes_regex: str | None = None
    inoopa_best_nace_codes: list[str] | None = None
    inoopa_best_nace_codes_regex: str | None = None

    minimum_number_of_estanlishments: int | None = None
    maximum_number_of_estanlishments: int | None = None
    employee_categories: list[EmployeeCategory] | None = None
    created_before: datetime | None = None
    created_after: datetime | None = None
    include_decision_makers: bool = False

    max_results: int | None = None
    excluded_companies: list[str] | None = None

    legal_form_types: list[LegalFormType] | None = None
    legal_forms: list[LegalForm] | None = None

    def to_mongo_filters(
        self,
        only_include_companies_with_phone_no_do_not_call_me: bool = False,
        only_include_companies_with_email: bool = False,
        only_include_companies_with_website: bool = False,
        only_include_companies_with_nace_codes: bool = False,
        only_include_active_companies: bool = True,
    ) -> MongoFilters:
        """
        Generate a MongoDB filter to search for companies based on the LeadGenerationCompanyFilters.

        can be used as a filter parameter in the MongoDB Collection.find() method.
        """
        filters = {}
        if self.countries:
            filters["country"] = {"$in": self.countries}
        if self.legal_forms:
            filters["legal_form"] = {"$in": self.legal_forms}
        if self.employee_categories:
            filters["employee_category_formatted"] = {"$in": self.employee_categories}
        if self.regions:
            regions = [r if r != "NOT FOUND" else None for r in self.regions]
            filters["address.region"] = {"$in": regions}
        if self.provinces:
            provinces = [p if p != "NOT FOUND" else None for p in self.provinces]
            filters["address.province"] = {"$in": provinces}
        if self.zipcodes:
            zipcodes = [z if z != "NOT FOUND" else None for z in self.zipcodes]
            filters["address.postal_code"] = {"$in": zipcodes}

        if self.minimum_number_of_estanlishments and self.maximum_number_of_estanlishments:
            filters["number_of_establishments"] = {
                "$gte": self.minimum_number_of_estanlishments,
                "$lte": self.maximum_number_of_estanlishments,
            }
        elif self.minimum_number_of_estanlishments:
            filters["number_of_establishments"] = {"$gte": self.minimum_number_of_estanlishments}
        elif self.maximum_number_of_estanlishments:
            filters["number_of_establishments"] = {"$lte": self.maximum_number_of_estanlishments}

        if self.declared_best_nace_codes:
            filters["$or"] = [
                {"nace_codes": {"$elemMatch": {"number": {"$in": self.declared_best_nace_codes}}}},
                {
                    "establishments": {
                        "$elemMatch": {"nace_codes": {"$elemMatch": {"number": {"$in": self.declared_best_nace_codes}}}}
                    }
                },
            ]
        if self.declared_best_nace_codes_regex:
            nace_section_codes = []
            naces_codes_regex = []
            for code_regex in self.declared_best_nace_codes_regex.split("|"):
                # if the regex is only one letter, it's a nace section code
                if len(code_regex) == 1:
                    nace_section_codes.append(code_regex)
                else:
                    # if the regex is more than one letter, it's a nace code regex
                    naces_codes_regex.append(code_regex)

            filters["$or"] = []
            if naces_codes_regex:
                filters["$or"].append(
                    {"nace_codes": {"$elemMatch": {"number": {"$regex": "|".join(naces_codes_regex)}}}}
                )
                filters["$or"].append(
                    {
                        "establishments": {
                            "$elemMatch": {
                                "nace_codes": {"$elemMatch": {"number": {"$regex": "|".join(naces_codes_regex)}}}
                            }
                        }
                    },
                )
            if nace_section_codes:
                filters["$or"].append({"nace_codes": {"$elemMatch": {"section_code": {"$in": nace_section_codes}}}})
                filters["$or"].append(
                    {
                        "establishments": {
                            "$elemMatch": {"nace_codes": {"$elemMatch": {"section_code": {"$in": nace_section_codes}}}}
                        }
                    },
                )

        if self.inoopa_best_nace_codes:
            filters["best_nace_codes.first_best_nace_code.number"] = {"$in": self.inoopa_best_nace_codes}
        if self.inoopa_best_nace_codes_regex:
            nace_codes = []
            nace_section_codes = []
            for code_regex in self.inoopa_best_nace_codes_regex.split("|"):
                # if the regex is only one letter, it's a nace section code
                if len(code_regex) == 1:
                    nace_section_codes.append(code_regex)
                else:
                    # if the regex is more than one letter, it's a nace code regex
                    nace_codes.append(code_regex)
            filters["$or"] = []
            if nace_codes:
                filters["$or"].append({"best_nace_codes.first_best_nace_code.number": {"$regex": "|".join(nace_codes)}})
            if nace_section_codes:
                filters["$or"].append(
                    {"best_nace_codes.first_best_nace_code.section_code": {"$in": nace_section_codes}}
                )

        if self.legal_form_types:
            filters["legal_form_type"] = {"$in": self.legal_form_types}
        if self.created_before and not self.created_after:
            filters["start_date"] = {"$lte": self.created_before}
        if self.created_after and not self.created_before:
            filters["start_date"] = {"$gte": self.created_after}
        if self.created_before and self.created_after:
            filters["start_date"] = {"$lte": self.created_before, "$gte": self.created_after}
        if self.excluded_companies:
            filters["_id"] = {"$nin": self.excluded_companies}
        if only_include_active_companies:
            filters["status"] = "Active"
            filters["legal_situation"] = "Normal situation"

        if only_include_companies_with_phone_no_do_not_call_me:
            filters["best_phone"] = {"$ne": None}
            filters["best_phone.phone"] = {"$nin": ["DO_NOT_CALL_ME", None]}
        if only_include_companies_with_email:
            filters["best_email"] = {"$ne": None}
        if only_include_companies_with_website:
            filters["best_website"] = {"$ne": None}
        if only_include_companies_with_nace_codes:
            filters["best_nace_codes"] = {"$ne": None}
        return filters

    def to_dict(self) -> dict:
        data_dict = self.model_dump()
        # Convert datetime objects to isoformat for json serialization
        if self.created_before:
            data_dict["created_before"] = _datetime_serializer(self.created_before)
        if self.created_after:
            data_dict["created_after"] = _datetime_serializer(self.created_after)
        return data_dict


class EnrichmentCompanyFilters(BaseModel):
    vats_to_enrich: list[str]
    country: Country = "BE"
    additional_fields: list[AdditionalFields] = ["email", "phone", "website", "nace_codes", "social_medias"]
    include_decision_makers: bool = False

    def to_mongo_filters(
        self,
        only_include_companies_with_phone_no_do_not_call_me: bool = False,
        only_include_companies_with_email: bool = False,
        only_include_companies_with_website: bool = False,
        only_include_companies_with_nace_codes: bool = False,
    ) -> MongoFilters:
        filters = {"country": self.country, "_id": {"$in": self.vats_to_enrich}}
        if only_include_companies_with_phone_no_do_not_call_me:
            filters["best_phone"] = {"$ne": None}
        if only_include_companies_with_email:
            filters["best_email"] = {"$ne": None}
        if only_include_companies_with_website:
            filters["best_website"] = {"$ne": None}
        if only_include_companies_with_nace_codes:
            filters["best_nace_codes"] = {"$ne": None}
        return filters

    def to_dict(self) -> dict:
        return self.model_dump()


class SemanticSearchCompanyFilters(BaseModel):
    countries: list[Country] = all_countries
    regions: list[Region] | None = None
    zipcodes: list[str] | None = None
    declared_best_nace_codes: list[str] | None = None
    declared_best_nace_codes_regex: str | None = None
    inoopa_best_nace_codes: list[str] | None = None
    inoopa_best_nace_codes_regex: str | None = None

    minimum_number_of_estanlishments: int | None = None
    maximum_number_of_estanlishments: int | None = None
    employee_categories: list[EmployeeCategory] | None = None
    created_before: datetime | None = None
    created_after: datetime | None = None
    include_decision_makers: bool = False

    max_results: int | None = None
    excluded_companies: list[str] | None = None

    legal_form_types: list[LegalFormType] | None = None
    legal_forms: list[LegalForm] | None = None

    additional_fields: list[AdditionalFields] | None = None

    def to_dict(self) -> dict:
        data_dict = self.model_dump()
        # Convert datetime objects to isoformat for json serialization
        if self.created_before:
            data_dict["created_before"] = _datetime_serializer(self.created_before)
        if self.created_after:
            data_dict["created_after"] = _datetime_serializer(self.created_after)
        return data_dict


def _datetime_serializer(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")



import os
from dotenv import load_dotenv
import pandas as pd
from pymongo.errors import OperationFailure
from pymongo import MongoClient, ASCENDING, TEXT
from concurrent.futures import ThreadPoolExecutor

from inoopa_utils.custom_types.decision_makers import DecisionMaker
from inoopa_utils.inoopa_logging import create_logger

load_dotenv()


class DbManagerMongo:
    """
    This class is used to manage the Mongo database.

    :param mongo_uri: The URI of the Mongo database to connect to.
    :param create_index_if_not_done: If the MongoDB indexes should be created if they don't exist.

    :attribute company_collection: The company collection object.
    :attribute do_not_call_me_collection: The do not call me collection object.
    :attribute decision_maker_collection: The decision maker collection object.
    :attribute legacy_co2_data_collection: The legacy co2 data collection object.
    :attribute legacy_decision_makers_collection: The legacy decision makers collection object.
    :attribute company_keywords: The company keywords collection object.

    :method update_do_not_call_me: Update the do_not_call_me collection in the database with a list of phone numbers.
    """

    def __init__(
        self,
        mongo_uri: str = os.environ["MONGO_READWRITE_PROD_URI"],
        create_index_if_not_done: bool = False,
        pretty_logging: bool = False,
    ):
        self._logger = create_logger("INOOPA_UTILS.DB_MANAGER.MONGO", pretty=pretty_logging)
        self._env = os.environ.get("ENV", "dev")

        _client = MongoClient(mongo_uri)
        _db = _client[self._env]

        self.company_collection = _db.get_collection("company")
        self.do_not_call_me_collection = _db.get_collection("do_not_call_me")
        self.decision_maker_collection = _db.get_collection("decision_maker")
        self.delivery_memory_collection = _db.get_collection("delivery_memory")
        self.legacy_co2_data_collection = _db.get_collection("legacy_co2_data")
        self.legacy_decision_makers_collection = _db.get_collection("legacy_decision_makers")
        self.phone_operators_cache_collection = _db.get_collection("phone_operators_cache")
        self.api_users_collection = _db.get_collection("api_users")
        self.company_keywords = _db.get_collection("company_keywords")
        if create_index_if_not_done:
            self._create_indexes()

    def update_do_not_call_me(self, phones: list[str]) -> None:
        """
        Method to update the do_not_call_me collection in the database.

        :param list[str]: The list of phone numbers to add to the do_not_call_me collection.
        """
        batch_size = 10_000
        phones_batch = [phones[i : i + batch_size] for i in range(0, len(phones), batch_size)]

        # We use a thread pool to update the database faster
        with ThreadPoolExecutor() as executor:
            executor.map(self._run_batch_update_do_not_call_me, phones_batch)
        self._logger.info(f"Updated {len(phones):_} phones in DoNotCallMe collection")

    def _run_batch_update_do_not_call_me(self, phone_batch: list[str]) -> None:
        """
        Method to update a chunk of the do_not_call_me collection in the database.

        :param updates: The list of DB update to execute.
        """
        batch_set = set(phone_batch)
        self._logger.info(f"Filtering {len(phone_batch)} phones...")
        existing_numbers = set(
            item["phone"] for item in self.do_not_call_me_collection.find({"phone": {"$in": list(batch_set)}})
        )
        new_numbers = [{"phone": number} for number in batch_set - existing_numbers]
        # Bulk insert new numbers
        if new_numbers:
            self._logger.info(f"Inserting {len(new_numbers)} new phones in collection...")
            self.do_not_call_me_collection.insert_many(new_numbers, ordered=False)
        else:
            self._logger.info("No new phones to insert found in batch")

    def _create_indexes(self) -> None:
        """Create the indexes in the Mongo database if they don't exist."""
        fields_to_index_legacy_co2_data = [
            [("_id", ASCENDING)],
            [("legacy_entity_id", ASCENDING)],
        ]
        fields_to_index_decision_makers = [
            [("company_id", ASCENDING)],
            [("legacy_entity_id", ASCENDING)],
            [("email", ASCENDING)],
            [("best_match", ASCENDING)],
            [("function_string", ASCENDING)],
            [("linkedin_url", ASCENDING)],
        ]
        fields_to_index_company = [
            [("address.string_address", ASCENDING)],
            [("address.region", ASCENDING)],
            [("address.postal_code", ASCENDING)],
            [("legal_form_type", ASCENDING)],
            [("best_website", ASCENDING)],
            [("best_website.website", ASCENDING)],
            [("best_email", ASCENDING)],
            [("best_phone", ASCENDING)],
            [("board_members.name", ASCENDING)],
            [("country", ASCENDING)],
            [("employee_category_code", ASCENDING)],
            [("establishments.name", ASCENDING)],
            [("name", ASCENDING)],
            [("status", ASCENDING)],
            [("name_text", TEXT)],
        ]
        legacy_fields_to_index_decision_makers = [
            [("company_id", ASCENDING)],
            [("email_score", ASCENDING)],
            [("phone_score", ASCENDING)],
            [("function_code", ASCENDING)],
            [("cluster", ASCENDING)],
            [("cluster_score", ASCENDING)],
            [("cluster_best_match", ASCENDING)],
            [("function_string", ASCENDING)],
        ]
        fields_to_index_company_keywords = [
            [("company_id", ASCENDING)],
        ]
        fields_to_index_delivery_memory = [
            [("company_id", ASCENDING)],
            [("best_phone.phone", ASCENDING)],
            [("best_email.email", ASCENDING)],
            [("best_website.website", ASCENDING)],
            [("delivery_date", ASCENDING)],
            [("decision_maker_ids", ASCENDING)],
        ]
        fields_to_index_phone_operators_cache = [
            [("operator", ASCENDING)],
            [("operator_last_update", ASCENDING)],
            [("last_update", ASCENDING)],
            [("_id", ASCENDING)],
        ]
        fields_to_index_do_no_call_me = [
            [("phone", ASCENDING)],
        ]
        self._logger.info("Creating indexes in collections (if not created yet)...")
        self._logger.info(
            "If this takes time (~10 mins), the indexes are not created yet. You will get a message when it's done."
        )

        collections_to_index = [
            {self.company_collection: fields_to_index_company},
            {self.decision_maker_collection: fields_to_index_decision_makers},
            {self.legacy_co2_data_collection: fields_to_index_legacy_co2_data},
            {self.legacy_decision_makers_collection: legacy_fields_to_index_decision_makers},
            {self.company_keywords: fields_to_index_company_keywords},
            {self.delivery_memory_collection: fields_to_index_delivery_memory},
            {self.do_not_call_me_collection: fields_to_index_do_no_call_me},
            {self.phone_operators_cache_collection: fields_to_index_phone_operators_cache},
        ]

        for indexer in collections_to_index:
            for collection, fields in indexer.items():
                for index in fields:
                    self._logger.debug(f"Creating index {index} in `{collection.name}` collection...")
                    try:
                        if index[0][0] != "_id":
                            collection.create_index(index, background=True)
                        else:
                            collection.create_index(index)
                    # Skip if index already exists
                    except OperationFailure as ex:
                        print(ex)
                        continue
        self._logger.info("Indexes created in all collections")

    def add_decision_makers_to_company(self, company_id: str) -> list[DecisionMaker]:
        """
        Add decision makers to companies list of dicts.

        :param companies: List of dicts containing companies data.
            Each company should contains a key `_id` or `registered_number`.
        :return: the list of decision makers objects.
        """
        self._logger.debug(f"Adding decision makers to company: {company_id}")
        decision_makers = self.decision_maker_collection.find({"company_id": company_id})
        if decision_makers:
            decision_makers = [DecisionMaker(**decision_maker) for decision_maker in decision_makers]
            self._logger.info(f"Found {len(decision_makers)} decision makers for company: {company_id}")
            return decision_makers

        self._logger.debug(f"No decision makers found for company: {company_id}")
        return []


def enrich_companies_df_with_decision_makers(
    companies_df: pd.DataFrame, db_manager: DbManagerMongo | None = None
) -> pd.DataFrame:
    """
    Add decision makers to a companies dataframe.

    :param companies_df: The companies dataframe to enrich. Should contain a column `__id` or `registered_number`.
    :return: The companies dataframe with the decision makers added.
        If multiple decision makers are found for one company, will create multiple rows for the company.
    """
    if companies_df.empty:
        return companies_df
    if db_manager is None:
        db_manager = DbManagerMongo()
    db_manager._logger.info("Adding decision makers to companies dataframe...")
    if "_id" in companies_df.columns:
        id_column = "_id"
    elif "registered_number" in companies_df.columns:
        id_column = "registered_number"
    else:
        raise ValueError("The companies dataframe should contain a column `_id` or `registered_number`.")

    for i, row in companies_df.iterrows():
        decision_makers = db_manager.add_decision_makers_to_company(row[id_column])
        if not decision_makers:
            continue
        else:
            # If there are multiple decision makers, create a new row for each one
            for i, decision_maker in enumerate(decision_makers):
                new_row = row.copy()
                new_row["decision_maker_firstname"] = decision_maker.firstname
                new_row["decision_maker_lastname"] = decision_maker.lastname
                new_row["decision_maker_email"] = decision_maker.email
                new_row["decision_maker_language"] = decision_maker.language
                new_row["decision_maker_function_cluster"] = decision_maker.cluster
                if i == 0:
                    companies_df.at[i] = new_row
                else:
                    companies_df = pd.concat([companies_df, pd.DataFrame([new_row])], ignore_index=True)
    # make sure the companies are displayed in the same order as the original dataframe
    return companies_df.sort_values(by="registered_number")


def enrich_companies_dicts_with_decision_makers(
    companies_dicts: list[dict], db_manager: DbManagerMongo | None = None
) -> list[dict]:
    """
    Add decision makers to a companies list of dicts.

    :param companies_dicts: List of dicts containing companies data.
        Each company should contains a key `_id` or `registered_number`.
    :return: the list of companies dicts with the decision makers added under the `decision_makers` key.
    """
    if db_manager is None:
        db_manager = DbManagerMongo()
    db_manager._logger.info("Adding decision makers to companies dicts...")
    for i, company_dict in enumerate(companies_dicts):
        decision_makers = db_manager.add_decision_makers_to_company(company_dict["_id"])
        companies_dicts[i]["decision_makers"] = []

        if not decision_makers:
            continue

        for dm in decision_makers:
            companies_dicts[i]["decision_makers"].append(
                {
                    "firstname": dm.firstname,
                    "lastname": dm.lastname,
                    "email": dm.email,
                    "language": dm.language,
                    "function_cluster": dm.cluster,
                }
            )

    return companies_dicts


if __name__ == "__main__":
    os.environ["ENV"] = "prod"
    db_manager = DbManagerMongo(create_index_if_not_done=True)
